module Simulation.Agents.ChemoTaxisAgent(generateChemoTaxisAgent) where

import Control.Monad.Random(getRandomR)
import Data.Function(on)
import Data.List(sortBy)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.Vec2
import qualified Simulation.Utility.Sampling as Sample

import Debug.Trace

generateChemoTaxisAgent numM len school starts sp num = CTA $
	map (\start -> createTaxisAgent numM len school start sp) starts

createTaxisAgent numM len school start sp = 
	let 	si  = createSimpleAgent' start sp
		sa  = Sample.createSamplingAvg numM
		sa2 = Sample.createSamplingAvg numM
	in TA si sa sa2 0 school (0, start) len

data ChemoTaxisAgent = CTA {
	cAgents :: ![TaxisAgent]
} deriving (Eq, Show)

data TaxisAgent = TA {
	simple		:: !SimpleAgent,		--Underlying agent
	samples		:: !(Sample.SamplingAvg Double),--Samples collected
	oldSamples	:: !(Sample.SamplingAvg Double),--Previous collected
	direction	:: !Float,			--Current travel direction
	useSchool	:: !Bool,			--Should schooling tactic be used?
	bestMeasure	:: (Double, Vec2 Float),	--Best measurement and position
	strideLength	:: !Float			--How far to move forward
} deriving (Eq, Show)

instance Drawable ChemoTaxisAgent where
	draw = Pictures . map draw . cAgents

instance Drawable TaxisAgent where
	draw = draw . simple

instance Agent ChemoTaxisAgent where
	position   = map (agentPos . simple) . cAgents
	totalLen   = sum . map (agentTravleLen . simple) . cAgents
	updateAgent step world transmitter agent = do
		let agents = cAgents agent
		newAgents  <- mapM (updateCTAAgent step world transmitter agents) agents
		return agent{cAgents = newAgents}

updateCTAAgent step world transmitter agents agent = do
	let update= if not $ useSchool agent
			then updateTaxisAgent agent agents
			else schoolUpdate agent agents
	newSimple<- moveSimpleAgent update step world (simple agent)
	let pos   = agentPos $ simple agent
	signal   <- signalAt transmitter pos
	let best  = compareBestMeasurement agent pos
	if agentDest (simple agent) /= agentDest newSimple
	then 	let new = Sample.createSamplingAvg (Sample.desiredCount $ samples agent)
	     	    old = samples agent
		    (x, y) = vec2Sub (agentDest newSimple) (agentPos newSimple)
		    newDir = atan2 y x
		in return agent{simple = newSimple, samples = new, oldSamples = old,
			direction = newDir, bestMeasure = best}
	else 	let new = Sample.addSample (samples agent) signal
		    old = oldSamples agent
		in return agent{simple = newSimple, samples = new, oldSamples = old,
			bestMeasure = best}

compareBestMeasurement agent position =
	if Sample.isReady (samples agent) && currentSamples  > bestSignal
	then (currentSamples, position)
	else bestMeasure agent
	where (bestSignal, _) = bestMeasure agent
	      currentSamples  = Sample.avg $ samples agent

getRandomDirection currentDir rand = do
	let r     = rand / 2
	addition <- getRandomR (-r, r)
	let new   = currentDir + addition
	return $ vec2Normalize (cos new, sin new)

closeWorldsEdge threshold (a, b) (a2, b2) = 
	let	wx = a/2
		wy = b/2
		outsideEast = wx - a2 <= threshold
		outsideWest = wx + a2 <= threshold
		outsideNorth= wy - b2 <= threshold
		outsideSouth= wy + b2 <= threshold
	in outsideEast || outsideWest || outsideNorth || outsideSouth

atWorldsEnd world agent = 
	let	currentPos = agentPos $ simple agent
		dir	   = direction agent
		currentDir = (cos dir, sin dir)
		closeToEdge= closeWorldsEdge 5 (worldSize world) currentPos
		headingOut = not $ insideWorld world $ vec2Add currentPos $
			--Add (10, 10) to ensure that the direction is outside the world
			--if the direction is pointing out
			vec2Scale 10 currentDir
	in closeToEdge && headingOut

updateTaxisAgent agent agents world simpleAgent = do
	if not $ Sample.isReady $ samples agent	--If not enough samples has been taken
	then return $ agentDest simpleAgent	--wait for more samples
	else if atWorldsEnd world agent
	then 	let 	currentDir = direction agent
			updateDir  = if currentDir < pi	
				    then currentDir + pi
				    else currentDir - pi
		in updateTaxisAgent agent{direction=updateDir} agents world simpleAgent
	else do
		let oldAvg = Sample.avg $ oldSamples agent
		let avg    = Sample.avg $ samples agent
		let rand   = if avg > oldAvg
				then 5   * (pi / 180)	--Convert to radians because 
				else pi 		--getRandomDirection works in radians
		newDir 	  <- getRandomDirection (direction agent) rand
		let scaled = vec2Scale (strideLength agent) newDir
		let target = vec2Add (agentPos simpleAgent) scaled
		return target

schoolUpdate agent agents world simpleAgent = do
	wantedMove    <- updateTaxisAgent agent agents world simpleAgent
	if wantedMove == agentDest simpleAgent
	then return wantedMove
	else do
		--Take the 3 closes neighbors as influences
		let influences= take 5 $ sortBy
			(compare `on`
			(distance (agentPos simpleAgent) . agentPos . simple)) agents
		let currentPos = agentPos simpleAgent
		let wanted'    = vec2Sub wantedMove currentPos
		let separation = calculateSeparation agent influences 150
		let cohesion   = calculateCohesion agent influences 250
		let dir        = calculateDirection influences
		let target     = calculateSchoolTarget agent wanted' dir separation cohesion
		let newTarget  = vec2Add currentPos target
		let turn       = vec2Scale (-1) target
		return $ if insideWorld world newTarget
			then newTarget
			--Move away from walls
			else vec2Add currentPos $ turn

--Calculate the separation so that agents are not on top of each other
calculateSeparation agent agents threshold = 
	let	positions = map
			(\a -> vec2Sub (agentPos $ simple a) (agentPos $ simple agent))
			agents
		close     = filter
			(\a -> let len = vec2Len a in len > 0 && len <= threshold)
			positions
		sep       = vec2Scale (-1) $ vec2Sum close
	in sep

--Calculate the cohesion an agent should have towards the school
calculateCohesion agent agents dist = 
	let	center   = vec2Sub centroid pos
	--If the agent is above a certain threshold from the center move towards
	--center if not continue
	in if distance (agentPos $ simple agent) centroid > dist then center else (0, 0)
	where 	centroid = calculateCentroid agents
		pos	 = agentPos $ simple agent

--Calculate the centroid of the school, used in cohesion calculations
calculateCentroid agents = vec2Scale (1 / n) $ vec2Sum positions
	where	positions= map (agentPos . simple) agents
		n	 = fromIntegral $ length agents

--Calculate the average direction in the school
calculateDirection agents = sum directions / n
	where 	directions= map direction agents
		n	  = fromIntegral $ length directions

--Calculate the direction to move towards given a desired direction to align to
--a maximum offset from that alignment and the desired vector to move to
calculateMaxDirection dir maxi (x, y) = max (dir - maxi') $ min (dir + maxi') wantedDir
	where	wantedDir = atan2 y x
		maxi'	  = maxi / 2

--Calculate the vector the agents wants to move to
--then control that its within some measure of the average direction
--the school wants to move towards
calculateSchoolTarget agent wanted dir sep coh = 
	let	moveTo = vec2Sum [vec2Scale 2 wanted, sep, coh]
		phi    = calculateMaxDirection dir (180 * (pi / 180)) moveTo
		scaled = vec2Scale (strideLength agent) $ vec2Normalize (cos phi, sin phi)
	in scaled
