module Simulation.Agents.EstimationAgent(generateEstimationAgent) where

import Control.Monad(liftM)
import GHC.Float(float2Double, double2Float)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.Vec2
import qualified Simulation.Utility.Sampling as Sample

import Debug.Trace(traceShow)

generateEstimationAgent numM numS start sp num = EA [EstiAgent simpleA numS sampling]
	where 	sampling = Sample.createSamplingAvg numM
		simpleA  = createSimpleAgent' (vec2Add (0.001, 0) start) sp

type EstimationSample = Sample.SamplingAvg (Vec2 Float, Double)

data EstimationAgent = EA {
	eAgents :: [EstiAgent]
} deriving (Eq, Show)

data EstiAgent = EstiAgent {
	simple  :: SimpleAgent,		--Underlying agent used for movement
	numSteps:: Double,		--Number of steps to take when moving
	samples :: EstimationSample	--Samples of position and RSS
} deriving (Eq, Show)

instance Drawable EstiAgent where
	draw = draw . simple

instance Drawable EstimationAgent where
	draw agent = Pictures $ map draw $ eAgents agent

instance Agent EstimationAgent where
	position   = map (agentPos . simple) . eAgents
	totalLen   = sum . map (agentTravleLen . simple) . eAgents
	updateAgent step world transmitter agent = do
		let agents = eAgents agent
		newAgents  <- mapM (updateEsAgent step world transmitter agents) agents
		return agent{eAgents = newAgents}

updateEsAgent step world transmitter agents agent = do
	newSimple <- moveSimpleAgent (updateEAgent agent agents) step world (simple agent)
	let pos   = agentPos $ simple agent
	signal    <- liftM convertAmplitudeToDB $ signalAt transmitter pos
	let newSamples = if agentDest (simple agent) /= agentDest newSimple
		then Sample.clearSamples $ Sample.addSample (samples agent) (pos, signal)
		else Sample.addSample (samples agent) (pos, signal)
	return agent{simple = newSimple, samples = newSamples}

updateEAgent agent agents world simpleAgent = do
	if not $ Sample.isReady $ samples agent	--If not enough samples has been taken
	then return $ agentDest simpleAgent	--wait for more samples
	else do
		let estimated = estimateRSSGradientDir (samples agent)
		let normEstim = vec2Normalize estimated
		let (x, y)    = vec2Scale (numSteps agent) normEstim
		let dest      = vec2Add (double2Float x, double2Float y) $
			agentPos simpleAgent
		return dest

convertAmplitudeToDB val = 20 * logBase 10 (val / 500)

estimateRSSGradientDir :: EstimationSample -> Vec2 Double
estimateRSSGradientDir sample = vec2Scale (1 / diver) (d * x1 - b * y1, -b * x1 + a * y1)
	where 	s = Sample.getSamples sample
		a = float2Double $ foldr (\x acc -> acc + x ** 2) 0 $ map (fst . fst) s
		d = float2Double $ foldr (\y acc -> acc + y ** 2) 0 $ map (snd . fst) s
		b = float2Double $
			foldr (\(x, y) acc -> acc + x * y) 0 $ map fst s
		(x1, y1) = foldr (\((x, y), z) (xs, ys) ->
			(xs + float2Double x * z, ys + float2Double y * z))
			(0, 0) s
		diver    = a * d - b ** 2
