{-# LANGUAGE Rank2Types, BangPatterns #-}
module Simulation.Agents.LevyRandomAgent3(generateLevyRandomAgent3) where

import Control.Monad(liftM)
import Control.Monad.Random(getRandomR)
import Data.Foldable(foldr')
import Data.Function(on)
import Data.List(sortBy)
import GHC.Float(double2Float, float2Double)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.LevyRandomAgent2(levyRandom, checkBounds)
import Simulation.Agents.SimpleAgent
import Simulation.Common

generateLevyRandomAgent3 alpha scale threshold pos sp = LRA (createSimpleAgent pos pos sp) alpha scale threshold False []

data PowerMeasurement = PM{
	measurementPos 	:: !(Float, Float),
	receivedPower 	:: !Double
} deriving (Eq, Show)

data LevyRandomAgent3 = LRA {
	agent 	  	:: !SimpleAgent,
	levyAlpha 	:: !Double,
	levyScale 	:: !Double,
	longThreshold 	:: !Double,
	record		:: !Bool,
	measurements	:: ![PowerMeasurement]
} deriving (Eq, Show)

changedGoal :: SimpleAgent -> SimpleAgent -> Bool
changedGoal a b = agentDest a /= agentDest b

type Vec2d = (Float, Float)

addVecs :: Vec2d -> Vec2d -> Vec2d
addVecs (!x1, !y1) (!x2, !y2) = (x1+x2, y1+y2)

subVecs a (!x, !y) = addVecs a (-x, -y)

scaleVec :: Float -> Vec2d -> Vec2d
scaleVec a (!x, !y) = (x*a, y*a)

-- |Calculate the estimated transmitter position from a list of sorted measurements
-- |this methods assumes a sorted list, splits the list in two, and creates a new list
-- |of n/2 size where it takes the largest vectors - the smallest vectors then it will
-- |add these vectors and scale it, the scaling might not be necessary
averagePosition sorted = averageVec
	where 	!(small, large)	= splitAt (round $ fromIntegral (length sorted) / 2) sorted
		!newVecs@(x:xs)	= concatMap (\v -> map (subVecs v) small) large
		!condensedVec	= foldr' addVecs x xs
		!averageVec	= flip scaleVec condensedVec $ 
					1 / fromIntegral (length newVecs)

estimateTransmitterDirection :: [PowerMeasurement] -> (Float, Float) -> Float
estimateTransmitterDirection samples pos = atan2 y x
	where 	!sorted = map measurementPos $ sortBy (compare `on` receivedPower) samples
		!estimated = averagePosition sorted
		(x, y) = subVecs estimated pos

selectNextPos :: LevyRandomAgent3 -> World -> SearchEnv ((Float, Float), Bool)
selectNextPos a w = do 
	let (x, y)   = agentPos $ agent a
	let (wX, wY) = worldSize w
	l'<- levyRandom $ levyAlpha a
	r <- getRandomR (0, 2*pi)
	let l = double2Float $ l' * levyScale a
	if l' > longThreshold a
	then 	if length (measurements a) < 2
		then return (checkBounds w (x + sin r * l, y + cos r * l), False)
		else do
			let nR = estimateTransmitterDirection (measurements a) (agentPos $ agent a)
			let next = (x + sin nR * l, y + cos nR * l)
			return (checkBounds w next, False)
	else return (checkBounds w (x + sin r * l, y + cos r * l), True)

moveLevyAgent step w t la@(LRA agent alpha scale thres currRec meas) = do 
	(nextPosition, rec) <- selectNextPos la w
	nAgent <- moveSimpleAgent (\_ _ -> return nextPosition) step w agent
	let recordNow = if changedGoal nAgent agent
			then rec
			else currRec
	if recordNow
	then do
		m <- receivedSignal t $ float2Double $
			distance (transmitterPos t) (agentPos agent)
		let m' = PM (agentPos agent) m
		return $ LRA nAgent alpha scale thres recordNow $ m' : meas
	else return $ LRA nAgent alpha scale thres recordNow []

instance Agent LevyRandomAgent3 where
	position a = [agentPos $ agent a]
	totalLen   = agentTravleLen . agent
	updateAgent= moveLevyAgent

instance Drawable LevyRandomAgent3 where
	draw la = Pictures [draw a, line points]
		where	a      = agent la
			points = map measurementPos $ measurements la
