{-# LANGUAGE Rank2Types #-}
module Simulation.Agents.ModPSO(generateModPSO) where

import Control.Monad(liftM)
import Data.Function(on)
import Data.List(maximumBy)
import GHC.Float(double2Float, float2Double)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.Vec2
import Simulation.Utility.Sampling

generateModPSO numMeasure thres moveMi moveMa starts sp num = ModPSO $ map
	(generateModPSOAgent numMeasure True thres moveMi moveMa sp) starts

generateModPSOAgent n scale thres moveMi moveMa sp start = ModPSOAgent (createSimpleAgent'
	start sp) scale thres
	moveMi moveMa $
	createSamplingAvg n

data ModPSO = ModPSO {
	mAgents :: ![ModPSOAgent]
} deriving (Eq, Show)

data ModPSOAgent = ModPSOAgent {
	simple 		:: !SimpleAgent,	--Underlying agent, helps with movement
	scaleM		:: !Bool,		--Should the vector between agents be scaled?
	threshold	:: !Double,		--Threshold used for SA approximation
	moveMin		:: !Double,		--Minimum movement, used to limit SA
	moveMax		:: !Double,		--Maximum movement
	samples		:: !(SamplingAvg Double) --Samples container
} deriving (Eq, Show)

instance Drawable ModPSO where
	draw (ModPSO ags) = Pictures $ map draw ags

instance Drawable ModPSOAgent where
	draw = draw . simple

instance Agent ModPSO where
	position   = map (agentPos . simple) . mAgents
	totalLen   = sum . map (agentTravleLen . simple) . mAgents
	updateAgent step world transmitter agent = do
		let agents = mAgents agent
		newAgents  <- mapM (updateModPSOAgent step world transmitter agents) agents
		return agent{mAgents = newAgents}

avgAmplitude :: Int -> ModPSOAgent -> (Double, ModPSOAgent)
avgAmplitude num agent = (avg $ samples agent, agent)

agentTravleVec :: (Double, ModPSOAgent) -> (Double, ModPSOAgent) -> Vec2 Float
agentTravleVec (aVal, aAgent) (bVal, bAgent) = 
	let scaleVal = if scaleM aAgent then double2Float $ abs aVal - bVal else 1.0
	in
		if aVal > bVal
		then vec2Scale scaleVal $ vec2Normalize $ vec2Sub aPos bPos
		else vec2Scale scaleVal $ vec2Normalize $ vec2Sub bPos aPos
	where   aPos = agentPos $ simple aAgent
		bPos = agentPos $ simple bAgent

createTravleVec :: (Double, ModPSOAgent) -> [(Double, ModPSOAgent)] -> Vec2 Float
createTravleVec comp = vec2Normalize . vec2Sum . map (agentTravleVec comp) 

addRandom :: Vec2 Float -> SearchEnv (Vec2 Float)
addRandom vec@(x, y) = do
	let len = vec2Len vec
	let phi = float2Double $ atan2 y x		--Angle in radians
	newPhi  <- liftM double2Float $ gaussian phi 1	--Add some stochasticity
	return (cos newPhi * len, sin newPhi * len)
	
updateModPSOAgent step world transmitter agents agent = do
	newSimple <- moveSimpleAgent (updateMPAgent agent agents) step world (simple agent)
	signal <- signalAt transmitter $ agentPos $ simple agent
	let newSamples = if agentDest (simple agent) /= agentDest newSimple
		then clearSamples $ addSample (samples agent) signal
		else addSample (samples agent) signal
	return agent{simple = newSimple, samples = newSamples}

calculateAvoidance agent agents = 
	--Filter agents closer than 50 which one wants to avoid
	let close = filter (\a ->
		let d = distance (agentPos $ simple agent) (agentPos $ simple a)
		in  d <= 50) agents
	--Added small amount in case two agents are on top of each other
	    sep   = map (\a -> vec2Sub (vec2Add (0.1, 0.1) $ agentPos $ simple a) (agentPos $ simple agent)) close
	in vec2Sum sep

--Calculate the cohesion an agent should have towards the school
calculateCohesion agent agents = 
	let	center   = vec2Sub centroid pos
	--If the agent is above a certain threshold from the center move towards
	--center if not continue
	in if distance (agentPos $ simple agent) centroid > 100 then center else (0, 0)
	where 	centroid = calculateCentroid agents
		pos	 = agentPos $ simple agent

--Calculate the centroid of the school, used in cohesion calculations
calculateCentroid agents = vec2Scale (1 / n) $ vec2Sum positions
	where	positions= map (agentPos . simple) agents
		n	 = fromIntegral $ length agents

updateMPAgent agent agents world simpleAgent = do
	if not $ isReady $ samples agent	--If not enough samples has been taken
	then return $ agentDest simpleAgent	--wait for more samples
	else do
		let n      = desiredCount $ samples agent
		let ownAvg = avgAmplitude n agent
		let others = map (avgAmplitude n) agents
		let temp   = threshold agent - fst ownAvg	--Current temperature
		let desired= if threshold agent > 0
			then double2Float $
				min (moveMax agent) $ max (moveMin agent) (temp * 10)
			else ((agentSpeed $ simple agent) / 60) * (fromIntegral n)
		let best   = maximumBy (compare `on` fst) others
		let ownVec = createTravleVec ownAvg others
		let bestVec= createTravleVec best others
		let avoid  = calculateAvoidance agent $ filter (/= agent) agents
		let center = calculateCohesion agent agents
		let sumVec = vec2Sum [ownVec, bestVec, center]
		scaled    <- addRandom $ vec2Scale (desired / vec2Len sumVec) sumVec
		return $ vec2Add (agentPos $ simple agent) scaled
