module Simulation.Agents.OrganizedAgent(generateOrganizedAgent, OrganizedAgent) where

import Control.Monad(liftM)
import Graphics.Gloss.Data.Picture(Picture(..))
import Simulation.Agents.SimpleAgent
import Simulation.Common

generateOrganizedAgent :: 	Float 		-> 
				World 		-> 
				[(Float, Float)]-> 
				Float 		-> 
				Int		->
				OrganizedAgent
generateOrganizedAgent distance world starts speed num = 
	--Since the distance is assumed to be maximum distance, it must be divided
	--by sqrt 2 so that the furthest distance is between an agent and the corner
	--or between two agents taking a turn
	if (distance / sqrt 2) * fromIntegral num >= wX
	then generateOrganizedAgent'' distance world starts speed num
	else generateOrganizedAgent' distance world starts speed num
	where	(wX, _) = worldSize world

--This version is used if the area is so small that a normal sweep along
--the Y axis is enough to cover the area
generateOrganizedAgent'' distance world starts' speed num = OA agents
	where	(wX, wY)	 = worldSize world
		sDist		 = distance / sqrt 2
		endY		 = wY / 2 - sDist / 2
		simpleAgents 	 = map (\start -> createSimpleAgent' start speed ) starts'
		starts 	= [((-wX/2) + sDist/2 + sDist * fromIntegral i, (-wY/2) + sDist/2) 
			| i <- [0..num-1]]
		paths		 = map (\pos@(x1, y1) -> [pos, (x1, endY)]) starts
		agents		 = zip (map cycle paths) simpleAgents

generateOrganizedAgent' distance world starts' speed num = OA agents
	where 	simpleAgents 	 = map (\start -> createSimpleAgent' start speed ) starts'
		(wX, wY) 	 = worldSize world
		--Maximum distance between two agents, xDist/2 max distance between
		--agent and wall
		xDist  		 = distance / sqrt 2
		paths  		 = map (calculatePath num wY xDist) [0..num-1]
		starts 	= [((-wX/2) + xDist/2 + xDist * fromIntegral i, (-wY/2) + xDist/2) 
			| i <- [0..num-1]]
		points 		 = map (\(start, path) -> 
			scanl convertPathToPoint start (path ++ cycle (tail path))) $ 
			zip starts paths
		numPoints 	 = map (length . 
			takeWhile (\(x, y) -> x <= wX/2 && y <= wY/2)) points
		points'   	 = map (\(n, p) -> take (n+1) p) $ zip numPoints points
		points''  	 = map (noSideBySide . map (pointInside (wX, wY))) points'
		points''' 	 = flip map points'' (\a ->
				if length a > 2
				then a ++ (tail $ init $ reverse a)
				else a ++ reverse a)
		agents    	 = zip (map cycle points''') simpleAgents

data PathDir = FirstUp | VerticalUp | VerticalDown | 
	HorizontalHigh | HorizontalLow deriving (Eq, Show)
type Path = (PathDir, Float)

pointInside (wX, wY) (x, y) = (nX, nY)
	where	nX = max (-wX/2) $ min (wX/2) x
		nY = max (-wY/2) $ min (wY/2) y

noSideBySide a = foldr (\n acc -> if head acc /= n then n:acc else acc) [last a] a

originPath = [FirstUp, HorizontalHigh, VerticalDown, HorizontalLow, VerticalUp]

cPath :: Int -> Float -> Float -> Int -> PathDir -> Float
cPath tAgs _ xDist cNum HorizontalHigh =
	fromIntegral (tAgs - cNum - 1) * xDist * 2 + xDist
cPath tAgs _ xDist cNum HorizontalLow  = fromIntegral cNum * xDist * 2 + xDist
cPath tAgs wY xDist cNum FirstUp  	    = wY - xDist - fromIntegral cNum * xDist
cPath tAgs wY xDist _ vert 	    = wY - xDist - fromIntegral (tAgs-1) * xDist

calculatePath :: Int -> Float -> Float -> Int -> [Path]
calculatePath tAgs wY xDist cNum = zip originPath distances
	where distances = map (cPath tAgs wY xDist cNum) originPath

convertPathToPoint :: (Float, Float) -> Path -> (Float, Float)
convertPathToPoint (x, y) (path, distance) = case path of
	HorizontalHigh 	-> (x+distance, y)
	HorizontalLow  	-> (x+distance, y)
	VerticalUp	-> (x, y+distance)
	VerticalDown	-> (x, y-distance)
	FirstUp		-> (x, y+distance)

data OrganizedAgent = OA {
	agents		:: [([(Float, Float)], SimpleAgent)]
} deriving (Eq, Show)

instance Drawable OrganizedAgent where
	draw ag = Pictures $ map (draw . snd) $ agents ag

moveSAgent step world ag@(path, a) = let
	dest = if null path then agentDest a else head path
	in do
		updated <- moveSimpleAgent (\_ _ -> return $ dest) step world a
		let nPath = if agentDest a /= agentDest updated 
			then tail path
			else path
		return (nPath, updated)

moveAgents step world _ (OA ags) = do
	newAgents <- mapM (moveSAgent step world) ags
	return $ OA newAgents

instance Agent OrganizedAgent where
	position ag = map (agentPos . snd) $ agents ag
	totalLen ag = sum $ map (agentTravleLen . snd) $ agents ag
	updateAgent = moveAgents
