module Simulation.Agents.RFMapAgent where

import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.GridMap
import Simulation.Utility.Vec2
import qualified Simulation.Utility.Sampling as S

data RFMapAgent = RFMapAgent {
	bbgf	:: !GridMap Double,	-- ^The binary, bayesian grid filter
	_sigma	:: !Double,		-- ^The noise on the receiver
	_agents	:: ![RFAgent]		-- ^The minions of this agent
}

data RFAgent = RFAgent {
	simple	:: !SimpleAgent,
	samples :: !(S.SamplingAvg Double)
}

instance Drawable RFMapAgent where
	draw (ModPSO ags) = Pictures $ map draw ags

instance Drawable RFAgent where
	draw = draw . simple

instance Agent RFMapAgent where
	position   = map (agentPos . simple) . _agents
	totalLen   = sum . map (agentTravleLen . simple) . _agents
	updateAgent step world transmitter agent = do
		let agents 	= _agents agent
		let entropyMap 	= convertWith logOddsToEntr $ bbgf agent
		newAgents  <- mapM 
			(updateRFAgent step world transmitter agents entropyMap) agents
		let samplesReady = filter (S.isReady . samples) newAgents
		let samplesPos   = map (\ag -> 
			let pos  = agentPos $ simple ag
			    pos' = convertToGridPos world pos $ bbgf agent
			    avg  = S.avg $ samples ag
			    avg' = convertAmplitudeToOdds (bbgf agent) (_sigma agent) avg
			in  (pos', avg')) samplesReady
		let newGrid      = updateGrid samplesPos $ bbgf agent
		return agent{_agents = newAgents, bbgf = newGrid}

--Used to convert log odds to entropy derived from paper
logOddsToEntr :: Double -> Double
logOddsToEntr logOdd = undefined

updateGrid :: [(Vec2 Int, Double)] -> GridMap Double -> GridMap Double
updateGrid pos grid = updateGrid grid pos

--This method converts a position in the agent world into a position in the grid map
--used by the agents to collect data. This function assumes that the (0, 0) in the grid map
--is in the lower left corner and that (world size, world size) is in the upper right.
convertToGridPos :: World -> Vec2 Float -> GridMap Double -> Vec2 Int
convertToGridPos world pos grid = convertToGridPos' wSize pos gSize
	where wSize = fst $ worldSize world
	      gSize = size grid

convertToGridPos' w (x, y) numCell = 
	let x'    = x + w/2
	    y'    = y + w/2
	    sSize = w / numCell
	in (floor $ x' / sSize, floor $ y' / sSize)
{-# INLINE convertToGridPos' #-}

convertAmplitudeToOdds :: GridMap Double -> Double -> Double -> Double
convertAmplitudeToOdds grid sigma amplitude = 
	--0.0 is used because there is no reflection in the environment
	let p     = amplitudeProbability sigma 0.0 amplitude
	    prior = 1 / ((size grid) ** 2)
	in log (1/pe - 1) - log (1/p - 1)

amplitudeProbability sigma l amplitude =
	let first  = 1 / (sqrt $ 2*pi * (sigma**2))
	    second = exp $ -0.5 * ((amplitude - l) ** 2)
	in first * second
{-# INLINE amplitudeProbability #-}
