module Simulation.Agents.RandomDirected(generateRandomDirAgent) where

import Control.Monad(liftM)
import Control.Monad.Random(getRandomR)
import Data.Function(on)
import Data.List(maximumBy)
import GHC.Float(double2Float, float2Double)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.Vec2
import qualified Simulation.Utility.Sampling as S

generateRandomDirAgent samples ticks threshold radian starts sp num =
	RandDirAgent $ map (generateRandAgent samples ticks threshold radian sp) starts

generateRandAgent samples ticks threshold radian sp start =
	RandAgent simple sample ticks 0 threshold radian
	where simple = createSimpleAgent' start sp
	      sample = S.createSamplingAvg samples

data RandDirAgent = RandDirAgent {
	_agents	:: ![RandAgent]		-- ^The minions of this agent
} deriving (Eq)

data RandAgent = RandAgent {
	simple		:: !SimpleAgent,
	samples 	:: !(S.SamplingAvg Double),
	ticks		:: !Int,
	currentTicks	:: !Int,
	threshold	:: !Double,
	rand		:: !Double 			--Random addition to add to movement
							--measured in radians
} deriving (Eq)

instance Drawable RandDirAgent where
	draw = Pictures . (map draw) . _agents

instance Drawable RandAgent where
	draw = draw . simple

instance Agent RandDirAgent where
	position   = map (agentPos . simple) . _agents
	totalLen   = sum . map (agentTravleLen . simple) . _agents
	updateAgent step world transmitter agent = do
		let agents 	= _agents agent
		newAgents  <- mapM 
			(updateRandAgent step world transmitter agents) agents
		return agent{_agents = newAgents}

updateRandAgent step world transmitter agents agent = do
	let agentsAbove = filter (\a -> let avg = S.avg $ samples a
				       in S.isReady (samples a) && avg >= threshold agent)
				agents
	signal <- signalAt transmitter $ agentPos $ simple agent
	let newSample = S.addSample (samples agent) signal
	if length agentsAbove > 0 && currentTicks agent <= 0
	then do
		let pos    = agentPos $ simple agent
		let maxAg  = maximumBy (compare `on` (S.avg . samples)) agentsAbove
		let target = if agent /= maxAg
			then agentPos $ simple maxAg
			else agentDest $ simple agent
		target'   <- if agent /= maxAg
			then addRandom target $ rand agent
			else return target
		let vecTarg= scaleVec2World world $ vec2Sub target' pos
		let s = simple agent
		return $ agent{simple = s{agentDest = vecTarg}, currentTicks = ticks agent}
	else do
		newSimple <- moveSimpleAgent moveRandAgent step world $ simple agent
		return $ agent{simple = newSimple, samples = newSample,
			currentTicks = (currentTicks agent - 1)}

scaleVec2World :: World -> Vec2 Float -> Vec2 Float
scaleVec2World w vec    = vec2Scale maxLength vecNorm
	where vecNorm   = vec2Normalize vec
	      maxLength = sqrt $ wx ** 2 + wy **2
	      (wx, wy)  = worldSize w
	    

--addRandom :: Vec2 Float -> SearchEnv (Vec2 Float)
addRandom vec@(x, y) rand = do
	let len = vec2Len vec
	let phi = float2Double $ atan2 y x		  --Angle in radians
	newPhi  <- liftM double2Float $ gaussian phi rand --0.05 Add some stochasticity
	return (cos newPhi * len, sin newPhi * len)

moveRandAgent world simple = do
	let (x, y)   = agentPos simple
	let (wx, wy) = worldSize world
	r <- getRandomR (0, 2*pi)
	let v = sqrt $ wx**2 + wy ** 2
	let next = (x + cos r * v, y + sin r * v)
	return next
