module Simulation.Agents.RandomPSO(generateRandPSOAgent) where

import Control.Monad(liftM)
import Control.Monad.Random(getRandomR)
import Data.Function(on)
import Data.List(maximumBy)
import GHC.Float(double2Float, float2Double)
import Graphics.Gloss.Data.Picture
import Simulation.Agents.SimpleAgent
import Simulation.Common
import Simulation.Utility.Vec2
import qualified Simulation.Utility.Sampling as S

import Debug.Trace

generateRandPSOAgent samples ticks threshold radian len starts sp num =
	RandPSOAgent $ map (generateRandAgent samples ticks threshold radian len sp) starts

generateRandAgent samples ticks threshold radian len sp start =
	RandAgent simple sample ticks 0 threshold Nothing Nothing radian start len
	where simple = createSimpleAgent' start sp
	      sample = S.createSamplingAvg samples

data RandPSOAgent = RandPSOAgent {
	_agents	:: ![RandAgent]		-- ^The minions of this agent
} deriving (Eq)

data RandAgent = RandAgent {
	simple		:: !SimpleAgent,
	samples 	:: !(S.SamplingAvg Double),
	ticks		:: !Int,
	currentTicks	:: !Int,
	threshold	:: !Double,
	personalBest	:: Maybe (Double, Vec2 Float),
	globalBest	:: Maybe (Double, Vec2 Float),
	randElement	:: !Double,
	lastRandDest	:: !(Vec2 Float),
	psoStep		:: !Float
} deriving (Eq, Show)

instance Drawable RandPSOAgent where
	draw = Pictures . (map draw) . _agents

instance Drawable RandAgent where
	draw = draw . simple

instance Agent RandPSOAgent where
	position   = map (agentPos . simple) . _agents
	totalLen   = sum . map (agentTravleLen . simple) . _agents
	updateAgent step world transmitter agent = do
		let agents 	= _agents agent
		newAgents  <- mapM 
			(updateRandAgent step world transmitter agents) agents
		return agent{_agents = newAgents}

updateRandAgent step world transmitter agents agent = do
	let agentsAbove = filter (\a -> let avg = S.avg $ samples a
				       in S.isReady (samples a) && avg >= threshold agent)
				agents
	signal <- signalAt transmitter $ agentPos $ simple agent
	let newSample = S.addSample (samples agent) signal
	let currentSimple = simple agent
	--If any agent have a signal above recall threshold
	--Or if currently engaged in PSO behaviour
	if (length agentsAbove > 0 && currentTicks agent <= 0) ||
		(length agentsAbove > 0 && globalBest agent /= Nothing)
	then do
		--If globalBest is nothing this is the first act of the PSO style
		--algorithm
		let _simple = case globalBest agent of
			Nothing -> currentSimple{agentDest = (agentPos currentSimple)}
			_	-> currentSimple
		let maxAg   = maximumBy (compare `on` (S.avg . samples)) agentsAbove
		let newG    = compareBestPos maxAg $ globalBest agent
		let newP    = compareBestPos agent $ personalBest agent
		--The agent with the best outlook should be pushed by the swarm
		--the other agents should move towards the best agent, PSO style
		let movement= if maxAg == agent
			      then movePSORandAgent2 maxAg agent agents
			      else movePSORandAgent newG newP agent
		newSimple  <- moveSimpleAgent movement step world _simple
		return $ agent{simple = newSimple, currentTicks = ticks agent,
			samples = newSample, globalBest = Just newG,
			personalBest = Just newP}
	else do
		newSimple <- moveSimpleAgent (moveRandAgent agent)
			step world currentSimple
		return $ agent{simple = newSimple, samples = newSample,
			currentTicks = (currentTicks agent - 1),
			globalBest = Nothing, personalBest = Nothing,
			lastRandDest = agentDest newSimple}

compareBestPos agent Nothing = (S.avg $ samples agent, agentPos $ simple agent)
compareBestPos agent (Just (val, pos)) = (bestAvg, bestPos)
	where bestAvg = if val > oldVal then val else oldVal
	      bestPos = if val > oldVal then pos else oldPos
	      oldVal  = S.avg $ samples agent
	      oldPos  = agentPos $ simple agent

--Use the simple PSO technique
movePSORandAgent (_, gPos) (_, pPos) agent world simple = do
	let len  = psoStep agent
	let rand = randElement agent
	let cPos = agentPos simple
	--Move towards the personal best position, but add some stochasticity
	pNew      <- flip addRandom rand $ vec2Sub pPos cPos
	--Move towards the global best position, but add some stochasticity
	gNew      <- flip addRandom rand $ vec2Sub gPos cPos
	--Current target
	let t      = vec2Add pNew gNew
	--Scale target such that movement is not to long at a time
	let target = vec2Add cPos $ vec2Scale len $ vec2Normalize t
	return target

--Use same technique as ModPSO
movePSORandAgent2 best agent agents world simpleAgent = do
	if not $ S.isReady $ samples agent	--If not enough samples has been taken
	then return $ agentDest simpleAgent	--wait for more samples
	else do
		let ownAvg = S.avg $ samples agent
		let others = zip (map (S.avg . samples) agents) agents
		let ownVec = createTravleVec (ownAvg, agent) others
		let bestVec= createTravleVec (S.avg $ samples best, best) others
		let sumVec = vec2Sum [ownVec, bestVec]
		let scaled = vec2Scale (psoStep agent / vec2Len sumVec) sumVec
		randomized<- addRandom scaled $ randElement agent
		return $ vec2Add (agentPos $ simple agent) scaled

agentTravleVec (aVal, aAgent) (bVal, bAgent) = 
		if aVal > bVal
		then vec2Sub aPos bPos
		else vec2Sub bPos aPos
	where   aPos = agentPos $ simple aAgent
		bPos = agentPos $ simple bAgent

createTravleVec comp = vec2Sum . map (agentTravleVec comp) 

addRandom vec@(x, y) rand = do
	let len = vec2Len vec
	let phi = float2Double $ atan2 y x		  --Angle in radians
	newPhi  <- liftM double2Float $ gaussian phi rand --0.05 Add some stochasticity
	return (cos newPhi * len, sin newPhi * len)

moveRandAgent agent world simple =
	if globalBest agent /= Nothing		--Previous movement was PSO related
	then return $ traceShow agent $ lastRandDest agent	--Continue with last selected random dest
	else do
		let (x, y)   = agentPos simple
		let (wx, wy) = worldSize world
		r <- getRandomR (0, 2*pi)
		let v = sqrt $ wx**2 + wy ** 2
		let next = (x + cos r * v, y + sin r * v)
		return next
