{-# LANGUAGE Rank2Types #-}
module Simulation.Agents.SimplePSO(generateSimplePSOAgent, generateAdvPSOAgent) where

import Control.Monad(liftM)
import Control.Monad.Random(getRandomR, getRandom)
import Data.Function(on)
import Data.List(maximumBy)
import Data.Traversable(forM)
import Graphics.Gloss.Data.Color
import Graphics.Gloss.Data.Picture
import Simulation.Common
import Simulation.Utility.Vec2

import Debug.Trace

type Pos = (Float, Float)

data PSO = PSO {
	agents :: [PSOAgent]
} deriving (Eq, Show)

data PSOAgent = PSOA {
	agentId		:: Int,		--A unique id is needed since agents can occupy 
					--the same space
	agentPos 	:: Pos,		--Current Agent position
	prevPos		:: Pos,
	randomDest	:: Maybe Pos,	--Used by the advanced PSO agent to enable random
					--movement when not in range of anything
	prevMeasurements:: [Double],	--List of signal measurements
					--used by advanced to normalize noise
	prevAvg		:: Double,
	pBestVal	:: Double,	--The personal best value for this agent
	pBestPos	:: Pos,		--The personal best position related to value
	vMax		:: Float,	--Max velocity in both directions
	vX		:: Float,	--Velocity in X
	vY		:: Float	-- ----||---- Y
} deriving (Eq, Show)

newtype AdvPSO = APSO {
	aPSOAgents :: [PSOAgent]
} deriving (Eq, Show)

generateAdvPSOAgent :: Pos -> Float -> Int -> AdvPSO
generateAdvPSOAgent start sp = APSO . flip take ags
	where 	nSp = sqrt $ (sp ** 2) / 2
		ags = map (createPSOAgent start nSp) [1..]

createPSOAgent :: Pos -> Float -> Int -> PSOAgent
createPSOAgent start sp aId = PSOA aId start start Nothing [] 0 (-1000000) start sp 0 0
	where 	fId = fromIntegral aId

generateSimplePSOAgent :: [Pos] -> Float -> Int -> PSO
generateSimplePSOAgent starts sp = PSO . flip take ags
	where 	nSp = sqrt $ (sp ** 2) / 2
		ags = map (\(id, start) -> createPSOAgent start nSp id) $ zip [0..] starts

updateDirection :: Float		--Current Position in direction
		-> Float		--Velocity in direction
		-> Float		--Personal best in direction
		-> Float		--Global best in direction
		-> Float		--Max velocity
		-> SearchEnv Float	--Update in direction
updateDirection curr vel pBest gBest vMax = do
	ran1 <- getRandom
	ran2 <- getRandom
	let v = vel + ran1*(pBest - curr) + ran2*(gBest - curr)
	return $ signum v * min vMax (abs v)

updatePSOAgent ::  Float 		--Time delta
		-> World 		--World
		-> Double 		--Current received signal
		-> (Double, Pos) 	--Swarm best signal and position
		-> PSOAgent 		--Agent to update
		-> SearchEnv PSOAgent	--Updated Agent
updatePSOAgent tDiff w r (bR, bPos) agent = do
	newX <- updateDirection (fst $ agentPos agent) (vX agent) (fst $ pBestPos agent)
		(fst bPos) (vMax agent)
	newY <- updateDirection (snd $ agentPos agent) (vY agent) (snd $ pBestPos agent)
		(snd bPos) (vMax agent)
	let (x, y) = agentPos agent
	let (pBV, pBP) =if r > pBestVal agent
			then (r, agentPos agent)
			else (pBestVal agent, pBestPos agent)
	let newPos = checkBounds w (x + newX*tDiff, y + newY*tDiff)
	return $ agent{agentPos = newPos, pBestVal = pBV, pBestPos = pBP, vX = newX, 
		vY = newY}

updateAPSOAgent tDiff w r agents signals gBestVec agent = do
	let (x, y) 	  = agentPos agent
	let others 	  = map agentPos $ filter (\ag -> agentId ag /= agentId agent) agents
	let fieldForce    = calculateForce (x, y) others
	let currentMeasure= take 100 $ r : (prevMeasurements agent)
	let currentAvg	  = calcAvg currentMeasure
	let probably      = currentAvg - prevAvg agent > 0.02
	(next, randDest) <- if probably
			     then do
				new <- newDirectedPosition r agents signals gBestVec agent
				return (new, Nothing)
			     else do
				new' <- newRandomPosition w agent
				let new = vec2Sub new' (x, y)
				return (new , Just new')
	let (newX, newY)= 
		--traceShow (agentId agent, currentAvg, prevAvg agent, 
		--probably) $ 
		vec2Sum [fieldForce, next]
	let newX'       =	signum newX * vMax agent
	let newY'       =	signum newY * vMax agent
	let newPos'     =	checkBounds w (x + newX' * tDiff, y + newY' * tDiff)
	let (pBV, pBP)  = 	if r > pBestVal agent
				then (r, agentPos agent)
				else (pBestVal agent, pBestPos agent)
	return $ agent{agentPos = newPos', prevPos = (x, y), vX = newX', 
		vY = newY', pBestVal = pBV, pBestPos = pBP,
		prevMeasurements=currentMeasure, randomDest = randDest,
		prevAvg = currentAvg}

newDirectedPosition r agents signals gBestVec agent = do
	let ownVec = calculateTravleVector (calcAvg $ prevMeasurements agent, agent) signals
--	ran1 <- getRandom
--	ran2 <- getRandom
--	return $ vec2Add (vec2Scale 0.3 ownVec) (vec2Scale 0.5 gBestVec)
--	return gBestVec
	let best    = maximumBy (compare `on` pBestVal) agents
	let current = agentPos agent
	return $ vec2Sum [calculateBestVec agent current,
		calculateBestVec best current]

newRandomPosition w agent = do
	case randomDest agent of
		Just pos -> do
			let dist = vec2Len $ vec2Sub pos (agentPos agent)
			if  dist <= agentCloseThreshold
			then newRandomPosition w $ agent{randomDest = Nothing}
			else return pos
		Nothing	 -> do
			let (wX, wY) = worldSize w
			x <- getRandomR (-wX/2, wX/2)
			y <- getRandomR (-wY/2, wY/2)
			return (x, y)

calcAvg :: [Double] -> Double
calcAvg li = sum li / fromIntegral (length li)

drawPSOAgent col agent = Color col $ Translate x y $ Circle agentSize
	where (x, y)   = agentPos agent


comparePSOVec :: (Double, PSOAgent) -> (Double, PSOAgent) -> Vec2 Float
comparePSOVec (aVal, a) (bVal, b) =     if aVal >= bVal
					then vec2Sub aPos bPos
					else vec2Sub bPos aPos
				where 	aPos = agentPos a
					bPos = agentPos b

calculateTravleVector :: (Double, PSOAgent) -> [(Double, PSOAgent)] -> Vec2 Float
calculateTravleVector comp = vec2Sum . map (comparePSOVec comp)

calculateForceBetween :: Vec2 Float -> Vec2 Float -> Vec2 Float
calculateForceBetween aVec bVec = vec2Scale (1/(len**2)) $ vec2Scale (1/len) neg
	where	neg  = vec2Sub bVec aVec
		len' = vec2Len neg
		len  = if len' > 0 then len' else 1

calculateForce :: Vec2 Float -> [Vec2 Float] -> Vec2 Float
calculateForce agent = vec2Scale (-1000) . 
	vec2Sum . map (calculateForceBetween agent)

getBestValuePos :: PSOAgent -> (Double, Vec2 Float)
getBestValuePos a = (pBestVal a, pBestPos a)

--Calculate the vector towards the best position recorded for the given agent
calculateBestVec :: PSOAgent -> Vec2 Float -> Vec2 Float
calculateBestVec agent current = vec2Sub goal current
	where	(_, goal) = getBestValuePos agent

instance Drawable PSO where
	draw pso = Pictures [Pictures $ map (drawPSOAgent red) $ agents pso, 
			drawPSOAgent black best]
		where best = maximumBy (compare `on` pBestVal) $ agents pso

instance Agent PSO where
	position pso = map agentPos $ agents pso
	totalLen _   = 0
	updateAgent diff world transmitter pso = do
		let ags  = agents pso
		let best = maximumBy (compare `on` pBestVal) ags
		let gBest = getBestValuePos best
		newAgents <- forM ags (\ agent -> do
			signal <- signalAt transmitter $ agentPos agent
			updatePSOAgent diff world signal gBest agent)
		return $ PSO newAgents

instance Drawable AdvPSO where
	draw pso = Pictures [Pictures $ map (drawPSOAgent red) $ aPSOAgents pso,
			Color black $ Translate x y $ Circle agentSize]
		where 	best   =  maximumBy (compare `on` pBestVal) $ aPSOAgents pso
			(x, y) = pBestPos best

instance Agent AdvPSO where
	position pso = map agentPos $ aPSOAgents pso
	totalLen _ = 0
	updateAgent diff world transmitter pso = do
		let ags = aPSOAgents pso
		agentsSignal <- liftM (flip zip ags) $ 
			mapM (return . calcAvg . prevMeasurements) ags
		let best  = maximumBy (compare `on` fst) agentsSignal
		let gBest = calculateTravleVector best agentsSignal
		newAgents <- forM ags (\agent -> do
			signal <- signalAt transmitter $ agentPos agent
			updateAPSOAgent diff world signal ags agentsSignal gBest agent)
		return $ APSO newAgents
		
