{-# LANGUAGE Rank2Types #-}
module Simulation.Common where

import Control.Monad.Random(Rand, RandomGen, getRandom)
import GHC.Float(float2Double)
import Graphics.Gloss.Data.Picture

type SearchEnv a = (RandomGen g) => Rand g a

data World = World {
	worldSize  :: !(Float, Float)
} deriving (Eq, Show)

class Drawable a where
	draw :: a -> Picture

class (Eq a, Drawable a) => Agent a where
	position      :: a -> [(Float, Float)]
	totalLen      :: a -> Float
	updateAgent   :: (Transmitter t) => Float -> World -> t -> a -> SearchEnv a

class (Eq a, Drawable a) => Transmitter a where
	transmitterPos  :: a -> (Float, Float)
	receivedSignal 	:: a -> Double -> SearchEnv Double
	withinRange	:: (Agent b) => a -> b -> SearchEnv Bool
	updateTrans	:: Float -> World -> a -> SearchEnv a

signalAt :: (Transmitter t) => t -> (Float, Float) -> SearchEnv Double
signalAt t pos = do
	let dist = float2Double $ distance pos $ transmitterPos t
	receivedSignal t dist

gaussian :: Double -> Double -> SearchEnv Double
gaussian mu sigma = do
	r1 <- getRandom
	r2 <- getRandom
	return $ mu + sigma * sqrt (-2 * log r1) * cos (2 * pi * r2)
{-# INLINE gaussian #-}

checkBounds :: World -> (Float, Float) -> (Float, Float)
checkBounds w (x, y) = (max (-wX) $ min wX x, max (-wY) $ min wY y)
	where   (wX', wY') 	= worldSize w
		wX 		= wX' / 2 - 1
		wY 		= wY' / 2 - 1
{-# INLINE checkBounds#-}

insideWorld :: World -> (Float, Float) -> Bool
insideWorld w (x, y) = x <= wX && x >= (-wX) && y <= wY && y >= (-wY)
	where	(wX', wY') = worldSize w
		wX	   = wX' / 2
		wY	   = wY' / 2
{-# INLINE insideWorld#-}

agentSize :: Float
agentSize = 5

agentCloseThreshold :: Float
agentCloseThreshold = 1

transmitterSize :: Float
transmitterSize = 5

distance :: (Floating a) => (a, a) -> (a, a) -> a
distance (x, y) (x2, y2) = sqrt $ (x2 - x) ** 2 + (y2 - y) ** 2
