{-# LANGUAGE OverloadedStrings, FlexibleInstances #-}
module Simulation.Configuration.Parsing(getJson, ParsedConfig(..),
	ParsingTransmitter(..), ParsingAgent(..),
	OutputConfig(..), OutputDataTypes(..),
	OutputMethod(..)) where

import Control.Applicative((<*>), (<$>))
import Control.Monad(liftM)
import Data.Aeson
import Data.Aeson.Types(Parser)
import Data.Text (Text)
import Simulation.Common(World(..))
import qualified Data.ByteString.Lazy as B

data ParsingTransmitter = Simple Float | 
	Alpha {
		effect		:: Double,
		alpha		:: Double,
		sigma		:: Double,
		position	:: Maybe (Float, Float),
		thresholdDeg	:: Double
} deriving (Eq, Show)
data ParsingAgent = Random | Organized Float | Levy Double Double |
	LevyDir Double Double Double | SimplePSO | AdvPSO |
	ModPSO Int Double Double Double |
	EstimationAgent Int Double | ChemoTaxisAgent Int Float Bool | 
	RandomDirAgent Int Int Double Double |
	RandomPSOAgent Int Int Double Double Float deriving (Eq, Show)

data OutputDataTypes = SimulationSteps | LengthTraveled deriving (Eq, Show)
data OutputMethod    = OneOnOne | OneAgainstAll | RoundRobin | Mean |
	RelativeMean deriving (Eq, Show)

data OutputConfig = OC {
	experimentName 		:: String,
	experimentConfidence 	:: Double,
	outputData		:: [OutputDataTypes],
	outputMethod		:: OutputMethod,
	extractData		:: Maybe [Int]
}deriving (Eq, Show)

getJson :: FilePath -> IO (Either String ParsedConfig)
getJson path = liftM eitherDecode' (B.readFile path)

data ParsedConfig = C {
	seed		:: !Int,
	simulations	:: !Int,
	numAgents	:: ![Int],
	numTransmitters :: !Int,
	agentSpeed	:: !Float,
	world		:: !World,
	transmitter	:: !ParsingTransmitter,
	agents		:: ![ParsingAgent],
	outputConf	:: !OutputConfig
}

instance FromJSON ParsedConfig where
	parseJSON (Object v) = 
		C <$> v .: "seed"
		<*> v .: "simulations"
		<*> v .: "numberOfAgents"
		<*> v .: "numberOfTransmitters"
		<*> v .: "agentSpeed"
		<*> v .: "world"
		<*> v .: "transmitter"
		<*> v .: "agents"
		<*> v .: "output"
	parseJSON _ = fail "Could not parse the configuration object"

instance FromJSON World where
	parseJSON (Object v) = 	do
				x <- v .: "worldX"
				y <- v .: "worldY"
				return $ World (x, y)
	parseJSON _	     = fail "World does not seem to be an object"

instance FromJSON ParsingTransmitter where
	parseJSON (Object v) = parseJSONTransmitter v
	parseJSON _	     = fail "Could not parse transmitter"

instance FromJSON ParsingAgent where
	parseJSON (Object v) = parseJSONAgent v
	parseJSON _	     = fail "Could not parse agent configuration"

instance FromJSON OutputConfig where
	parseJSON (Object v) = 	OC
				<$> v .: "name"
				<*> v .: "confidence"
				<*> v .: "outputTypes"
				<*> v .: "outputMethod"
				<*> v .:? "extractData"
	parseJSON _	     = fail "Could not parse output config"

instance FromJSON OutputDataTypes where
	parseJSON (Object v) = do
		outputType <- v .: "type"
		case outputType of
			"SimulationSteps" -> return SimulationSteps
			"LengthTraveled"  -> return LengthTraveled
			_		  -> fail $ "Unknown output type: " ++ outputType
	parseJSON _ = fail "Could not understand output data types"

instance FromJSON OutputMethod where
	parseJSON (Object v) = do
		outputType <- v .: "type" :: Parser Text
		case outputType of
			"OneOnOne"      -> return OneOnOne
			"OneAgainstAll" -> return OneAgainstAll
			"RoundRobin"    -> return RoundRobin
			"Mean"		-> return Mean
			"RelativeMean"  -> return RelativeMean
			_		-> fail $ 
				"Unknown output method: " ++ show outputType
	parseJSON _ = fail "Could not understand output method selected"

parseJSONTransmitter :: Object -> Parser ParsingTransmitter
parseJSONTransmitter v = do
	transType <- v .: "type" :: Parser String
	case transType of
		"SimpleTransmitter" 	-> parseSimpleTransmitter v
		"AlphaTransmitter"	-> parseAlphaTransmitter v
		_			-> fail $ 
			"Unknown transmitter type: " ++ show transType

parseJSONAgent :: Object -> Parser ParsingAgent
parseJSONAgent v = do
	agentType <- v .: "type" :: Parser Text
	case agentType of
		"RandomAgent"	  -> parseRandomAgent v
		"LevyAgent"	  -> parseLevyAgent v
		"LevyDirAgent"    -> parseLevyDirAgent v
		"OrganizedAgent"  -> parseOrganizedAgent v
		"SimplePSO"	  -> parseSimplePSO v
		"AdvPSO"	  -> parseAdvPSO v
		"ModPSO"	  -> parseModPSO v
		"EstimationAgent" -> parseEstiAgent v
		"ChemoTaxisAgent" -> parseChemoAgent v
		"RandomDirAgent"  -> parseRandomDirAgent v
		"RandomPSOAgent"  -> parseRandomPSOAgent v
		_		  -> fail $ "Unknown agent type: " ++ show agentType

--Implementation specific parsing
parseSimpleTransmitter :: Object -> Parser ParsingTransmitter
parseSimpleTransmitter v = Simple
	<$> v .: "radius"

parseAlphaTransmitter :: Object -> Parser ParsingTransmitter
parseAlphaTransmitter v = Alpha
	<$> v .: "effect"
	<*> v .: "alpha"
	<*> v .: "sigma"
	<*> v .:? "position"
	<*> v .: "threshold"

parseRandomAgent _ = return Random

parseLevyAgent v = Levy
	<$> v .: "alpha" 
	<*> v .: "scale"

parseLevyDirAgent v = LevyDir
	<$> v .: "alpha"
	<*> v .: "scale"
	<*> v .: "threshold"

parseOrganizedAgent v = Organized <$> v .: "distance"

parseSimplePSO _ = return SimplePSO

parseAdvPSO _ = return AdvPSO

parseModPSO v = ModPSO 
	<$> v .: "numMeasurements"
	<*> v .: "threshold"
	<*> v .: "moveMin"
	<*> v .: "moveMax"

parseEstiAgent v = EstimationAgent 
	<$> v .: "numMeasurements"
	<*> v .: "numSteps"

parseChemoAgent v = ChemoTaxisAgent
	<$> v .: "numMeasurements"
	<*> v .: "stride"
	<*> v .: "school"

parseRandomDirAgent v = RandomDirAgent
	<$> v .: "numMeasurements"
	<*> v .: "ticks"
	<*> v .: "threshold"
	<*> v .: "radians"

parseRandomPSOAgent v = RandomPSOAgent
	<$> v .: "numMeasurements"
	<*> v .: "ticks"
	<*> v .: "threshold"
	<*> v .: "radians"
	<*> v .: "stepLength"
