module Simulation.Output(outputFromConfig) where

import Control.Monad(liftM, forM_, (=<<), when)
import Control.Monad.Reader(ReaderT, runReaderT, ask, reader)
import Control.Monad.Trans.Class(lift)
import Data.List(transpose, nub)
import Data.Time.Format(formatTime)
import Data.Time.LocalTime(getZonedTime)
import GHC.Float(float2Double)
import Simulation.Configuration.Parsing(OutputConfig(..), 
	OutputDataTypes(..), OutputMethod(..))
import Simulation.Graphics.Gnuplot as Plot
import Simulation.Model(SearchModelStatistics(..))
import Simulation.Runnable(ResultStatistics)
import Simulation.Statistics(twoSamplePairedComparison, twoSampleWelshComparison,
	simpleMean, relativeMean)
import System.FilePath(combine)
import System.Locale(defaultTimeLocale)
import Text.Printf(printf)
import qualified System.Directory as Dir

type OutputReader a = ReaderT OutputConfig IO a

resultFolderName = "results"

resultFolderPath :: OutputReader FilePath
resultFolderPath = do
	current <- lift Dir.getCurrentDirectory
	name    <- reader experimentName
	return $ current `combine` resultFolderName `combine` name

getResultTime :: IO String
getResultTime = do
	currentTime <- getZonedTime
	return $ formatTime defaultTimeLocale "%F %R" currentTime

resultFolderFilename :: String -> OutputReader FilePath
resultFolderFilename filename = do
	dir <- resultFolderPath
	return $ dir `combine` filename

initOutput :: OutputReader ()
initOutput = do
	dir <- resultFolderPath
	lift $ Dir.createDirectoryIfMissing True dir

copyExperimentFiles :: FilePath -> OutputReader ()
copyExperimentFiles configFile = do
	outputDir <- resultFolderPath
	lift $ Dir.copyFile configFile $ outputDir `combine` "config.json"

writeResultDataToFile :: Show a => FilePath -> a -> OutputReader ()
writeResultDataToFile filename dat = writeStringResultToFile filename $ show dat

writeStringResultToFile filename dat = do
	filepath <- resultFolderFilename filename
	lift $ writeFile (filepath ++ ".txt") dat

createOutput :: [(String, [ResultStatistics])] -> OutputReader ()
createOutput dat = do
	writeResultDataToFile "raw-data" dat
	outputFormats <- reader outputData
	outputM       <- reader outputMethod
	method        <- selectOutputMethod
	outDat	      <- modifyData dat
	when (length dat > 1 || outputM == Mean) $
		forM_ (nub outputFormats) (\ ot ->
			case ot of
			SimulationSteps -> method (fromIntegral . totalSteps) outDat
			LengthTraveled  -> method (float2Double . totalLength) outDat)

modifyData dat = do
	extract <- reader extractData
	case extract of
		Just nums -> return $ extractData' nums dat
		Nothing   -> return dat

extractData' nums dat = map (\(strat, n) -> (strat, 
	map (\(num, d) -> (num, map (\i -> d !! i) nums)) n)) dat

selectOutputMethod = do
	selectedMethod <- reader outputMethod
	case selectedMethod of
		RoundRobin    -> return outputRoundRobin
		OneOnOne      -> return outputOneOnOne
		OneAgainstAll -> return outputOneAgainstRest
		Mean	      -> return outputSimpleMean
		RelativeMean  -> return outputRelativeMean

outputSimpleMean f dat = do
	let title = "Mean data"
	let means = simpleMean f dat
	writeResultDataToFile title means

outputRelativeMean f dat = do
	let means     = relativeMean f dat
	let tableMean = showRelativeMean means
	writeResultDataToFile "Raw-relative-means" means
	writeStringResultToFile "Relative table" tableMean
	
showRelativeMean means = unlines $ map showStrategy means
	where showStrategy (name, def, rest) =
		let table      = unlines $ map showData rest
		    headerline = "table[y error=error]{"
		    endline    = "};"
		in unlines ["%" ++ name, headerline, "x y error", table, endline]
		where showData (x, (y, error)) = printf "%d %f %f" x y error

outputOneOnOne f dat = do
	let one = head dat
	let two = head $ tail dat
	confidence <- reader experimentConfidence
	let comparison = twoSamplePairedComparison f confidence one two
	let title = "Comparison of " ++ fst one ++ " & " ++ fst two
--	let xRange = zipWith (\x i -> (show $ fst x, i)) comparison [0..]
--	let oneData = (fst one, map (fst . snd) comparison)
--	let twoData = (fst two, map (snd . snd) comparison)
--	let frame   = Plot.histogram2d title xRange $ oneData : twoData : []
	filename <- resultFolderFilename title
	writeResultDataToFile title comparison
--	lift $ print =<< Plot.plotToScreen frame
--	lift $ print =<< Plot.plotToEps filename frame

outputOneAgainstRest f dat = do
	let one = head dat
	let title = "Comparison against " ++ fst one
	confidence' <- reader experimentConfidence
	let a = 1 - confidence'
	let confidence = 1 - (a / (fromIntegral (length dat) - 1))
	let comparisons = map (twoSamplePairedComparison f confidence one) $ tail dat
--	let xRange  = zipWith (\x i -> (show $ fst x, i)) (head comparisons) [0..]
--	let cp (x, y) = x - y
--	let newData = map (map (cp . snd)) comparisons
--	let names   = map fst $ tail dat
--	let dataPoints = zip names newData
--	let frame = Plot.histogram2d title xRange dataPoints
	filename  <- resultFolderFilename title
	writeResultDataToFile title comparisons
--	lift $ print =<< Plot.plotToScreen frame
--	lift $ print =<< Plot.plotToEps filename frame

outputRoundRobin f dat = do
	let title = "Comparison everyone against everyone"
	confidence' <- (reader experimentConfidence)
	let a = 1 - confidence'
	let len = fromIntegral $ length dat
	let confidence = 1 - (a / (len * (len - 1)/2))
	let comparisons = concat $ roundRobin (\one two -> 
		("Comparing " ++ fst one ++ " with " ++ fst two, 
		twoSamplePairedComparison f confidence one two)) dat
	writeResultDataToFile title comparisons
	let names = map fst comparisons
	let orderedByNumTrans' = transpose $ map snd comparisons
	let orderedByNumTrans = map (zip names) orderedByNumTrans'
	forM_ orderedByNumTrans (\dataPoints -> do
		let numTrans = fst $ snd $ head dataPoints
		writeResultDataToFile
			(title ++ " number of agents" ++ show numTrans) $
			dataPoints)
		

roundRobin :: (a -> a -> b) -> [a] -> [[b]]
roundRobin f (x:[]) = []
roundRobin f (x:xs) = map (f x) xs : roundRobin f xs

outputFromConfig :: FilePath -> OutputConfig -> [(String, [ResultStatistics])] -> IO ()
outputFromConfig file conf dat = do
	time <- getResultTime
	let conf' = conf{experimentName= time ++ " " ++ experimentName conf}
	runReaderT initOutput conf'
	flip runReaderT conf' $ copyExperimentFiles file
	flip runReaderT conf' $ createOutput dat
