module Simulation.Runnable(runSimulation, runGraphically, ResultStatistics) where

import Control.Concurrent(forkIO)
import Control.Concurrent.Async(mapConcurrently)
import Control.Concurrent.MVar(putMVar, takeMVar, newEmptyMVar, MVar)
import Control.DeepSeq(force)
import Control.Monad(liftM, liftM2, replicateM, forM)
import Control.Monad.Random(evalRand, getRandom, getRandomR, evalRandIO)
import Simulation.Agents.ChemoTaxisAgent(generateChemoTaxisAgent)
import Simulation.Agents.EstimationAgent(generateEstimationAgent)
import Simulation.Agents.LevyRandomAgent2(generateLevyRandomAgent2)
import Simulation.Agents.ModPSO(generateModPSO)
import Simulation.Agents.OrganizedAgent(generateOrganizedAgent)
import Simulation.Agents.RandomAgent(generateRandomAgent)
import Simulation.Agents.RandomDirected(generateRandomDirAgent)
import Simulation.Agents.RandomPSO(generateRandPSOAgent)
import Simulation.Agents.SimplePSO(generateSimplePSOAgent, generateAdvPSOAgent)
import Simulation.Common
import Simulation.Configuration.Parsing
import Simulation.Model(runSearchModel, createSearchModel, SearchModelStatistics(..), runSearchModelGraphically)
import Simulation.Transmitters.AlphaTransmitter(generateAlphaTransmitter, AlphaTransmitter(..))
import System.IO(hPutStr, stderr)
import System.Random(setStdGen, newStdGen, mkStdGen, StdGen, RandomGen)
import System.Random.TF.Init(newTFGen)

{-
 -The results coming from this module is modeled as such:
 -Type	 	Random			Levy		...
 -#Agents	|1|2|...|		|1|2|...|
 -Transmitter	| | |   |		| | |   |
 -Configuration:| | |   |		| | |   |
 ----------------------------------------------------------
 - #1		|[]|[]|[]...|		|[]|[]|[]...|
 - #2		|[]|[]|[]...|		|[]|[]|[]...|
 - ...
-}
type ResultStatistics = (Int, [[SearchModelStatistics]])

runSimulation :: ParsedConfig -> IO [(String, [ResultStatistics])]
runSimulation config = do
	let ags  	= numAgents config
	let sims 	= simulations config
	let numTrans	= numTransmitters config
	let w	 	= world config
	let (wX, wY) 	= worldSize w
	let sp   	= agentSpeed config
	let moveIn	= 0.05 * wX -- Done so the agents do not start in the extreme corner
	let start	= (-wX/2 + moveIn, -wY/2 + moveIn)
	startLoc	<- replicateM (maximum ags) $ selectRandomStart start moveIn
	--let (Simple d) 	= transmitter config
	let transCreator = extractTransmitter config
	counter		<- newEmptyMVar
	transmitters <- evalRandIO $ replicateM numTrans transCreator
	--print $ map transPos transmitters
	forkIO $ showProgress 0 (length (agents config)*length ags*sims*numTrans) counter
	flip mapConcurrently (agents config) (\ agent -> do
		case agent of
			SimplePSO -> do
				let gen = (\num st -> [generateSimplePSOAgent st sp num])
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			Random -> do
				let gen = \num starts -> map 
					(\st -> generateRandomAgent st sp)
					starts
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			(RandomDirAgent numM ticks thres radians) -> do
				let gen = \num st -> 
					[generateRandomDirAgent numM ticks thres radians
					st sp num]
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			(ChemoTaxisAgent numM len school) -> do
				let gen = 
					(\num st ->
					[generateChemoTaxisAgent numM len school st sp num])
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			(ModPSO numM thres mMin mMax) -> do
				let gen = 
					(\num st ->
					[generateModPSO numM thres mMin mMax st sp num])
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			(Organized d) -> do
				let gen = (\num st -> [generateOrganizedAgent d w 
					st sp num])
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			(Levy alpha scale) -> do
				let gen = \_ starts -> map 
					(\st -> generateLevyRandomAgent2 alpha scale st sp)
					starts
				r <- gatherResults ags w sims gen transmitters counter startLoc
				putStrLn $ "Agent: " ++ show agent ++ ", done!"
				return (show agent, r)
			_ -> undefined	--Very dangerous, as it will fail silently
		)

runGraphically :: ParsedConfig -> IO ()
runGraphically config = do
	let w	 	= world config
	let (wX, wY) 	= worldSize w
	let sp   	= agentSpeed config
	let moveIn	= 0.05 * wX -- Done so the agents do not start in the extreme corner
	let start	= (-wX/2 + moveIn, -wY/2 + moveIn)
	agent	     	<- selectFromMultiple "Agent configuration" $ agents config
	ags 	     	<- selectFromMultiple "Number of agents" $ numAgents config
	startLoc 	<- replicateM ags $ selectRandomStart start moveIn
	t            	<- evalRandIO $ extractTransmitter config
	case agent of
		SimplePSO -> do
			let as = [generateSimplePSOAgent startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		Random -> do
			let as = map (\st -> generateRandomAgent st sp) startLoc
			runSearchModelGraphically $ createSearchModel w t as
		(RandomDirAgent numM ticks thres radians) -> do
			let as = [generateRandomDirAgent numM ticks thres radians
				startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		(RandomPSOAgent numM ticks thres radians len) -> do
			let as = [generateRandPSOAgent numM ticks thres radians len
				startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		(ModPSO numM thres mMin mMax) -> do
			let as = [generateModPSO numM thres mMin mMax startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		(ChemoTaxisAgent numM len school) -> do
			let as = [generateChemoTaxisAgent numM len school startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		(Organized d) -> do
			let as = [generateOrganizedAgent d w startLoc sp ags]
			runSearchModelGraphically $ createSearchModel w t as
		(Levy alpha scale) -> do
			let as = map
				(\start -> generateLevyRandomAgent2 alpha scale start sp)
				startLoc
			runSearchModelGraphically $ createSearchModel w t as

selectFromMultiple :: (Show a) => String -> [a] -> IO a
selectFromMultiple desc li = do
	let choiceStr = "There are multiple choices for " ++ desc
	putStrLn $ flip replicate '-' $ length choiceStr
	putStrLn choiceStr
	mapM_ putStrLn $ zipWith (\n i -> show n ++ ":\t" ++ show i) [0..] li
	num <- liftM read getLine
	if num >= 0 && num < length li
	then return $ li !! num
	else selectFromMultiple desc li

selectRandomStart (x, y) r = do
	dir <- getRandomR (0, 2*pi)
	len <- getRandomR (0, r)
	return (x + cos dir * len, y + sin dir * len)

extractTransmitter config = case (transmitter config) of
	(Alpha r a sigma mPos t) 	-> do
		let (wX, wY) = worldSize $ world config
		(x, y) <- case mPos of
			Just pos -> return pos
			Nothing  -> do
				x <- getRandomR (-wX/2, wX/2)
				y <- getRandomR (-wY/2, wY/2)
				return (x, y)
		return $ generateAlphaTransmitter (x, y) r a sigma t
	_		        -> error "Could not parse transmitter"

gatherResults ::	(Agent a, Transmitter t)    	=>
                        [Int]        			->
                        World        			->
			Int	     			->
                        (Int -> [(Float, Float)] -> [a]) ->
			[t] 	     			->
			MVar ()	     			->
			[(Float, Float)] 		->
                        IO [ResultStatistics]
gatherResults n w sims alg trans counter startLoc = do
        mapConcurrently (runAsync w sims alg trans counter startLoc) n

sndT (_, t, _) = t

runAsync ::     (Agent a, Transmitter t)             =>
                World                                ->
		Int				     ->
                (Int -> [(Float, Float)] -> [a])     ->
		[t]				     ->
		MVar ()				     ->
		[(Float, Float)]		     ->
                Int 				     ->
                IO ResultStatistics
runAsync w sims genAgents trans counter startLoc numAgs = do
	let starts = take numAgs startLoc
        let ags    = genAgents numAgs starts
	gen	  <- newTFGen
	let r   = flip evalRand gen $ mapM (\t ->
			replicateM sims $
			runSearchModel $ createSearchModel w t ags) trans
	!r' 	<- mapM (mapM (updateProgress counter)) r
        return (numAgs, r')

updateProgress :: MVar () -> SearchModelStatistics -> IO SearchModelStatistics
updateProgress counter result = do
	let !res = force result
	putMVar counter ()
	return res

showProgress :: Int -> Int -> MVar () -> IO ()
showProgress curr total var = do
	if curr == total then return () else do
		takeMVar var
		let curr' = curr + 1
		hPutStr stderr $ "\r" ++ (calculateProgress curr' total)
		showProgress curr' total var

calculateProgress :: Int -> Int -> String
calculateProgress curr total = show percentage ++
	"%(" ++ show curr ++ "/" ++ show total ++ ")"
	where percentage = floor $ (fromIntegral curr / fromIntegral total) * 100
