module Simulation.Statistics(simpleMean, 
			     twoSampleWelshComparison,
			     twoSamplePairedComparison,
			     relativeMean
			     ) where

import Simulation.Model(SearchModelStatistics)
import Statistics.Distribution(quantile, complCumulative)
import Statistics.Distribution.Binomial(binomial)
import Statistics.Distribution.StudentT(studentT)

tDistFromConf conf n = flip quantile conf $ studentT n

-- | relativeMean calculates the mean value for each strategy relative to 
-- the first result, this method is meant to show the improvements that arise
-- of using several agents
relativeMean :: (SearchModelStatistics -> Double) 		->
		[(String, [(Int, [[SearchModelStatistics]])])] 	->
		[(String, Int, [(Int, (Double, Double))])]
relativeMean f dat     = 
	let agents     = map fst dat				--Extract agents strings
	    runs       = map snd dat				--Extract actual data
	    numAgents  = map fst $ head runs			--Get number of agents
								--this will be equal for all
	    converted  = map (map (map (map f) . snd)) runs	--Convert data
	    avgPerTrans= map (map (map calcAvg)) converted
	    comparison = map head avgPerTrans
	    relative   = map (zipWith relativeZip comparison . tail) avgPerTrans
	    relativeAvg= map (map calcAvgStdev) relative
	in zipWith (\agent dat -> (agent, head numAgents, zip (tail numAgents) dat)) agents relativeAvg
	where relativeZip = zipWith (\comp a -> let b = abs $ a / comp
		in if isNaN b then 1 else b)

simpleMean ::	(SearchModelStatistics -> Double) -> 
		[(String, [(Int, [[SearchModelStatistics]])])] -> 
		[(String, [(Int, Double, Double)])]
simpleMean f dat 		     = zip agents results
	where	agents 		     = map fst dat
		runs   	    	     = map snd dat
		numAgents 	     = map fst $ head runs
		rawData		     = map (map (map f . concat . snd)) runs
		converted 	     = map (map calcAvgStdev) rawData
		results 	     = map (zipWith sowTogether numAgents) converted
		sowTogether a (b, c) = (a, b, c)

twoSampleGenericComparison f comp conf (_, first) (_, second) = zip numAgs intervals
	where	intervals = map compareConfidences $ 
			zipWith (\(_, dat1) (_, dat2) -> 
				comp f conf dat1 dat2) first second
		numAgs = map fst first

twoSampleWelshComparison ::  (SearchModelStatistics -> Double) ->
			Double ->
			(String, [(Int, [[SearchModelStatistics]])]) ->
			(String, [(Int, [[SearchModelStatistics]])]) ->
			[(Int, (Double, Double, Double))]
twoSampleWelshComparison f conf a b = twoSampleGenericComparison f pairedWelshComparison conf a b

twoSamplePairedComparison ::  (SearchModelStatistics -> Double) ->
			Double ->
			(String, [(Int, [[SearchModelStatistics]])]) ->
			(String, [(Int, [[SearchModelStatistics]])]) ->
			[(Int, (Double, Double, Double))]
twoSamplePairedComparison f conf a b = twoSampleGenericComparison f pairedTComparison conf a b

-- |This method compares the confidences in a list by performing a binomial test,
-- described in the thesis appendix B
compareConfidences :: [(Ordering, (Double, Double))] -> (Double, Double, Double)
compareConfidences dat = (complCumulative dist (first + second), first, second)
	where	first  = fromIntegral $ length $ filter ((==GT) . fst) dat
		second = fromIntegral $ length $ filter ((==LT) . fst) dat
		total  = fromIntegral $ length dat
		dist   = binomial total 0.5

-- |Do a paired comparison of two results, this method will calculate the mean and
-- |standard deviation for each:
--	TransmitterConfig:	| Statistics #1 | Statistics #2
--		1		| [...]		|	[...]
--		2               | [...]         |       [...]
--		3               | [...]         |       [...]
--		...             | ...           |       ...
--	Turned into:
--		1               | avg,stddev    |       avg, stddev
--		2               | avg,stddev    |       avg, stddev
--		3               | avg,stddev    |       avg, stddev
--	Using a t-test/two mean difference test assuming different
--	variance(unknown) using:
--
--	(x1, x2) - t_value * sqrt ( s1^2 / n + s2^2 / n) < u1 - u2 <
--	(x1, x2) + t_value * sqrt ( s1^2 / n + s2^2 / n)
--	where x1 is the mean for Statistics #1 and s1 is the standard deviation.
--	Page 251 Probability & Statistics - For engineers & scientists by
--	Walpole - Myers - Myers - YE. Seventh edition.
--
--	We then add an ordering which says if both intervals are > 0 then
--	that indicates that on the average #1 is better than #2
pairedWelshComparison :: (SearchModelStatistics -> Double) ->
			 Double ->
			 [[SearchModelStatistics]] ->
			 [[SearchModelStatistics]] ->
			 [(Ordering, (Double, Double))]
pairedWelshComparison f conf fir sec = map (\conf -> (confIntToOrdering conf, conf)) intervals
	where	firAvg = map (calcAvgStdev . map f) fir
		secAvg = map (calcAvgStdev . map f) sec
		intervals = zipWith (twoSampleWelshConfidenceInterval conf) firAvg secAvg

-- |Covert a confidence interval into an ordering when the numbers are assumed to
-- |originate from two means. Page 250 "Probability & Statistics for engineers & scientists"
confIntToOrdering :: (Double, Double) -> Ordering
confIntToOrdering (lower, upper)
	| lower > 0 && upper > 0  = LT
	| lower < 0 && upper < 0  = GT
	| otherwise		  = EQ

-- |Calculate the 95% confidence interval for two means with their standard deviation
-- |It is assumed that these means are calculated from a sufficiently large number
-- Also from Simulation modelling and statistics page. 554
twoSampleWelshConfidenceInterval :: Double -> 
				    (Double, Double) -> 
				    (Double, Double) -> 
				    (Double, Double)
twoSampleWelshConfidenceInterval conf (x1, s1) (x2, s2) = (left, right)
	where	left  = diff - confT*varianceSq
		right = diff + confT*varianceSq
		diff  = x1 - x2
		confT = tDistFromConf conf undefined
		varianceSq = sqrt $ (s1**2) / 2 + (s2**2) / 2

pairedTComparison :: (SearchModelStatistics -> Double)  ->
		     Double		 		->
		     [[SearchModelStatistics]] 		->
		     [[SearchModelStatistics]] 		->
		     [(Ordering, (Double, Double))]
pairedTComparison f conf dat1 dat2 = map (\conf -> (confIntToOrdering conf, conf)) intervals
	where	zs        = zipWith (zipWith (-)) convX convY
		convX     = map (map f) dat1
		convY     = map (map f) dat2
		intervals = map (twoSamplePairedTConf conf) zs

--From Simulation modelling and statistics page. 553
twoSamplePairedTConf :: Double -> [Double] -> (Double, Double)
twoSamplePairedTConf conf zs = (meanZ - rest, meanZ + rest)
	where	meanZ = sum zs / len
		len   = fromIntegral $ length zs
		rest  = confT * varianceSq
		confT = tDistFromConf conf $ fromIntegral (length zs) - 1
		varianceSq = sqrt variance
		variance   = foldr (\z acc -> (z - meanZ) ** 2 + acc) 0 zs
			/ (len * (len - 1))

calcAvgStdev :: [Double] -> (Double, Double)
calcAvgStdev li       = (avg, stdev)
	where	avg   = calcAvg li
		stdev = calcStdDev li avg

calcAvg :: [Double] -> Double
calcAvg li 	    = s / len
	where	s   = sum li
		len = fromIntegral $ length li

calcStdDev :: [Double] -> Double -> Double
calcStdDev li avg   = sqrt $ s / len
	where	s   = foldr (\n a -> a + (n - avg)**2) 0 li
		len = fromIntegral $ length li
